//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------

#ifndef FACTOR_GNSS_SINGLE_DIFF_2d_H_
#define FACTOR_GNSS_SINGLE_DIFF_2d_H_

//Wolf includes
#include "core/common/wolf.h"
#include "core/factor/factor_autodiff.h"
#include "core/frame/frame_base.h"
#include "gnss/sensor/sensor_gnss.h"

namespace wolf {
    
WOLF_PTR_TYPEDEFS(FactorGnssSingleDiff2d);

class FactorGnssSingleDiff2d : public FactorAutodiff<FactorGnssSingleDiff2d, 3, 2, 1, 2, 1, 3, 1, 1, 1>
{
    protected:
        SensorGnssPtr sensor_gnss_ptr_;

    public:

        FactorGnssSingleDiff2d(FeatureBasePtr& _ftr_ptr, const FrameBasePtr& _frame_other_ptr, const SensorGnssPtr& _sensor_gnss_ptr, const ProcessorBasePtr& _processor_ptr, bool _apply_loss_function, FactorStatus _status = FAC_ACTIVE) :
            FactorAutodiff<FactorGnssSingleDiff2d, 3, 2, 1, 2, 1, 3, 1, 1, 1>("GNSS SINGLE DIFFERENCES 2d",
                                                                              TOP_GEOM,
                                                                              _ftr_ptr,
                                                                              _frame_other_ptr,
                                                                              nullptr,
                                                                              nullptr,
                                                                              nullptr,
                                                                              _processor_ptr,
                                                                              _apply_loss_function,
                                                                              _status,
                                                                              _frame_other_ptr->getP(),
                                                                              _frame_other_ptr->getO(),
                                                                              _ftr_ptr->getFrame()->getP(),
                                                                              _ftr_ptr->getFrame()->getO(),
                                                                              _sensor_gnss_ptr->getP(),
                                                                              _sensor_gnss_ptr->getEnuMapRoll(),
                                                                              _sensor_gnss_ptr->getEnuMapPitch(),
                                                                              _sensor_gnss_ptr->getEnuMapYaw()),
            sensor_gnss_ptr_(_sensor_gnss_ptr)
        {
            WOLF_WARN_COND(!sensor_gnss_ptr_->isEnuDefined(), "Creating a GNSS SingleDiff 2d factor without initializing ENU");
        }

        virtual ~FactorGnssSingleDiff2d() = default;

        template<typename T>
        bool operator ()(const T* const _x1,
                         const T* const _o1,
                         const T* const _x2,
                         const T* const _o2,
                         const T* const _x_antena,
                         const T* const _roll_ENU_MAP,
                         const T* const _pitch_ENU_MAP,
                         const T* const _yaw_ENU_MAP,
                         T* _residuals) const;

};

template<typename T>
inline bool FactorGnssSingleDiff2d::operator ()(const T* const _x1,
                                                const T* const _o1,
                                                const T* const _x2,
                                                const T* const _o2,
                                                const T* const _x_antena,
                                                const T* const _roll_ENU_MAP,
                                                const T* const _pitch_ENU_MAP,
                                                const T* const _yaw_ENU_MAP,
                                                T* _residuals) const
{
    Eigen::Map<const Eigen::Matrix<T,2,1> > t_MAP_BASE1(_x1);
    Eigen::Matrix<T,2,2> R_MAP_BASE1 = Eigen::Rotation2D<T>(_o1[0]).matrix();
    Eigen::Map<const Eigen::Matrix<T,2,1> > t_MAP_BASE2(_x2);
    Eigen::Matrix<T,2,2> R_MAP_BASE2 = Eigen::Rotation2D<T>(_o2[0]).matrix();
    Eigen::Map<const Eigen::Matrix<T,3,1> > t_BASE_ANTENA(_x_antena);
    Eigen::Map<Eigen::Matrix<T,3,1> > residuals_ECEF(_residuals);

    Eigen::Matrix<T,3,3> R_ENU_ECEF = sensor_gnss_ptr_->getREnuEcef().cast<T>();
    Eigen::Matrix<T,2,3> R_MAP_ENU = sensor_gnss_ptr_->computeREnuMap(_roll_ENU_MAP[0], _pitch_ENU_MAP[0], _yaw_ENU_MAP[0]).transpose().topRows(2);

    // Transform ECEF 3d SingleDiff Feature to 2d SingleDiff in Map coordinates (removing z)
    Eigen::Matrix<T,2,1> measured_diff_2d_MAP = R_MAP_ENU * (R_ENU_ECEF * getMeasurement().cast<T>());

    // Substraction of expected antena positions in Map coordinates
    Eigen::Matrix<T,2,1> expected_diff_2d_MAP = (R_MAP_BASE2 * t_BASE_ANTENA.head(2) + t_MAP_BASE2) - (R_MAP_BASE1 * t_BASE_ANTENA.head(2) + t_MAP_BASE1);

    // Compute residual rotating information matrix to 2d Map coordinates
    // Covariance & information are rotated by R*S*R' so for the squred root upper (or right) R*L*U*R'->U*R'
    // In this case R = R_2dMAP_ENU * R_ENU_ECEF, then R' = R_ENU_ECEF' * R_2dMAP_ENU'
    residuals_ECEF = (getMeasurementSquareRootInformationUpper().cast<T>() * R_ENU_ECEF.transpose() * R_MAP_ENU.transpose()) * (expected_diff_2d_MAP - measured_diff_2d_MAP);

    //std::cout << "frame1: " << _x1[0] << " " << _x1[1] << " " << _o1[0] << std::endl;
    //std::cout << "frame2: " << _x2[0] << " " << _x2[1] << " " << _o2[0] << std::endl;
    //std::cout << "antena: " << _x_antena[0] << " " << _x_antena[1] << " " << _x_antena[2] << std::endl;
    //std::cout << "RPY: " << _roll[0] << " " << _pitch[1] << " " << _yaw[2] << std::endl;
    //std::cout << "measured_diff_2d: " << measured_diff_2d[0] << " " << measured_diff_2d[1] << std::endl;
    //std::cout << "expected_diff_2d: " << expected_diff_2d[0] << " " << expected_diff_2d[1] << std::endl;

    return true;
}

} // namespace wolf

#endif
